@@ -54,6 +54,15 @@ ResDecl eta_period   ("xfer:eta-period", "120",ResMgr::UNumberValidate,ResMgr::N
 ResDecl max_redir    ("xfer:max-redirections", "5",ResMgr::UNumberValidate,ResMgr::NoClosure);
 ResDecl buffer_size  ("xfer:buffer-size","0x10000",ResMgr::UNumberValidate,ResMgr::NoClosure);
 
+enum {
+   // Delays in microseconds
+   MAX_DELAY=30000,
+   DELAY_STEP=30,
+   // This size is related to socket buffer size.
+   // When it is too large, tcp slowdown happens.
+   MAX_READ_TO_DELAY=0x4000,
+};
+
 // FileCopy
 #define super SMTask
 
@@ -413,8 +422,8 @@ int FileCopy::Do()
 
 FileCopy::FileCopy(FileCopyPeer *s,FileCopyPeer *d,bool c)
    : get(s), put(d), cont(c),
-   rate(new Speedometer("xfer:rate-period")),
-   rate_for_eta(new Speedometer("xfer:eta-period"))
+   rate("xfer:rate-period"),
+   rate_for_eta("xfer:eta-period")
 {
    set_state(INITIAL);
    int max_buf=buffer_size.Query(0);
@@ -528,52 +537,52 @@ const char *FileCopy::GetPercentDoneStr() const
 }
 void FileCopy::RateAdd(int a)
 {
-   rate->Add(a);
-   rate_for_eta->Add(a);
+   rate.Add(a);
+   rate_for_eta.Add(a);
 }
 void FileCopy::RateReset()
 {
    start_time=now;
-   rate->Reset();
-   rate_for_eta->Reset();
+   rate.Reset();
+   rate_for_eta.Reset();
 }
-float FileCopy::GetRate() const
+float FileCopy::GetRate()
 {
-   if(!rate->Valid() || !put)
+   if(!rate.Valid() || !put)
       return 0;
-   return rate->Get();
+   return rate.Get();
 }
-const char *FileCopy::GetRateStr() const
+const char *FileCopy::GetRateStr()
 {
-   if(!rate->Valid() || !put)
+   if(!rate.Valid() || !put)
       return "";
-   return rate->GetStrS();
+   return rate.GetStrS();
 }
-off_t FileCopy::GetBytesRemaining() const
+off_t FileCopy::GetBytesRemaining()
 {
    if(!get)
       return 0;
    if(get->range_limit==FILE_END)
    {
       off_t size=get->GetSize();
-      if(size<=0 || size<get->GetRealPos() || !rate_for_eta->Valid())
+      if(size<=0 || size<get->GetRealPos() || !rate_for_eta.Valid())
 	 return -1;
       return(size-GetPos());
    }
    return get->range_limit-GetPos();
 }
-const char *FileCopy::GetETAStr() const
+const char *FileCopy::GetETAStr()
 {
    off_t b=GetBytesRemaining();
    if(b<0 || !put)
       return "";
-   return rate_for_eta->GetETAStrSFromSize(b);
+   return rate_for_eta.GetETAStrSFromSize(b);
 }
-long FileCopy::GetETA(off_t b) const
+long FileCopy::GetETA(off_t b)
 {
-   if(b<0 || !rate_for_eta->Valid())
+   if(b<0 || !rate_for_eta.Valid())
       return -1;
-   return (long)(double(b) / rate_for_eta->Get() + 0.5);
+   return (long)(double(b) / rate_for_eta.Get() + 0.5);
 }
 const char *FileCopy::GetStatus()
 {
@@ -638,6 +647,15 @@ void FileCopy::LogTransfer()
       const char *fname = ResMgr::Query("xfer:log-file", 0);
       if(!fname || !*fname)
 	 fname = dir_file(get_lftp_data_dir(),"transfer_log");
+      struct stat st;
+      if(stat(fname,&st)!=-1) {
+	 if(st.st_size > long(ResMgr::Query("xfer:max-log-size",0))) {
+	    debug((9,"rotating xfer-log %s\n",fname));
+	    if(rename(fname,xstring::cat(fname,".old",NULL))==-1) {
+	       debug((1,"rename(%s): %s\n",fname,strerror(errno)));
+	    }
+	 }
+      }
       int fd = open(fname,O_WRONLY|O_APPEND|O_CREAT,0600);
       if(fd==-1)
 	 return;
@@ -677,7 +695,7 @@ bool FileCopy::CheckFileSizeAtEOF() const
    const long long get_pos=get->GetRealPos();
    const long long put_pos=put->GetRealPos();
    const long long pos=(get_pos>put_pos ? get_pos : put_pos);
-   if(pos>=range_limit)
+   if(pos<=0 || pos>=range_limit)
       return true;
 
    debug((0,"expected pos=%lld, actual pos=%lld\n",range_limit,pos));
@@ -748,6 +766,35 @@ void FileCopyPeer::Seek(off_t offs)
    broken=false;
 }
 
+const char *FileCopyPeer::UseTempFile(const char *file)
+{
+   if(!ResMgr::QueryBool("xfer:use-temp-file",0))
+      return file;
+
+   xstring &temp=xstring::get_tmp(ResMgr::Query("xfer:temp-file-name",0));
+   if(temp.length()==0 || temp.eq("*"))
+      return file;
+
+   auto_rename=true;
+   temp_file=true;
+
+   const char *name=basename_ptr(file);
+   SetSuggestedFileName(name);
+
+   int subst_pos=temp.instr('*');
+   if(subst_pos>=0)
+      temp.set_substr(subst_pos,1,name);
+   else {
+      if(temp.last_char()=='.')
+	 temp.append(name);
+      else if(temp[0]=='.')
+	 temp.set_substr(0,0,name);
+      else
+	 temp.append('.').append(name);
+   }
+   return dir_file(dirname(file),temp);
+}
+
 FileCopyPeer::FileCopyPeer(dir_t m) : IOBuffer(m)
 {
    want_size=false;
@@ -767,6 +814,7 @@ FileCopyPeer::FileCopyPeer(dir_t m) : IOBuffer(m)
    range_limit=FILE_END;
    removing=false;
    file_removed=false;
+   temp_file=false;
    use_cache=true;
    write_allowed=true;
    done=false;
@@ -782,6 +830,22 @@ int FileCopyPeerFA::Do()
    int m=STALL;
    int res;
 
+   if(session->OpenMode()==FA::RENAME)
+   {
+      int res=session->Done();
+      if(res==FA::IN_PROGRESS)
+	 return m;
+      if(res<0) {
+	 if(temp_file)
+	    SetError(session->StrError(res));
+	 else
+	    debug((3,"rename failed: %s\n",session->StrError(res)));
+      }
+      session->Close();
+      done=true;
+      return MOVED;
+   }
+
    if(removing)
    {
       res=session->Done();
@@ -805,6 +869,12 @@ int FileCopyPeerFA::Do()
 	 SetError(verify->ErrorText());
       if(verify->Done())
       {
+	 if(suggested_filename && auto_rename)
+	 {
+	    const char *new_name=dir_file(dirname(file),suggested_filename);
+	    session->Rename(file,new_name);
+	    return MOVED;
+	 }
 	 done=true;
 	 m=MOVED;
       }
@@ -892,14 +962,7 @@ int FileCopyPeerFA::Do()
 		  return m;
 	       if(res==FA::STORE_FAILED)
 	       {
-		  try_time=session->GetTryTime();
-		  retries=session->GetRetries();
-		  off_t pos=session->GetRealPos();
-		  if(upload_watermark<pos) {
-		     upload_watermark=pos;
-		     retries=-1;
-		  }
-		  Log::global->Format(10,"try_time=%ld, retries=%d\n",(long)try_time,retries);
+		  upload_state.Save(session);
 		  session->Close();
 		  if(can_seek && seek_pos>0)
 		     Seek(FILE_END);
@@ -920,7 +983,7 @@ int FileCopyPeerFA::Do()
 	 return m;
       if(fxp)
 	 return m;
-      res=Get_LL(GET_BUFSIZE);
+      res=TuneGetSize(Get_LL(get_size));
       if(res>0)
       {
 	 EmbraceNewData(res);
@@ -1033,16 +1096,15 @@ void FileCopyPeerFA::OpenSession()
    session->Open(file,FAmode,seek_pos);
    session->SetFileURL(orig_url);
    session->SetLimit(range_limit);
-   if(mode==PUT)
-   {
-      if(try_time!=NO_DATE)
-	 session->SetTryTime(try_time);
-      if(retries>=0)
-	 session->SetRetries(retries+1);
+   if(mode==PUT) {
+      upload_state.Restore(session);
       if(e_size!=NO_SIZE && e_size!=NO_SIZE_YET)
 	 session->SetSize(e_size);
       if(date!=NO_DATE && date!=NO_DATE_YET)
 	 session->SetDate(date);
+   } else {
+      if(size!=NO_SIZE && size!=NO_SIZE_YET)
+	 session->SetSize(size);
    }
    session->RereadManual();
    if(ascii)
@@ -1075,6 +1137,13 @@ void FileCopyPeerFA::RemoveFile()
 
 int FileCopyPeerFA::Get_LL(int len)
 {
+   if(get_delay>0)
+   {
+      if(!get_ll_timer.Stopped())
+	 return 0;
+      session->ResumeSlave();
+   }
+
    int res=0;
 
    if(session->IsClosed())
@@ -1087,7 +1156,7 @@ int FileCopyPeerFA::Get_LL(int len)
    if(GetRealPos()!=io_at) // GetRealPos can alter pos.
       return 0;
 
-   res=session->Read(GetSpace(len),len);
+   res=session->Read(this,len);
    if(res<0)
    {
       if(res==FA::DO_AGAIN)
@@ -1109,7 +1178,7 @@ int FileCopyPeerFA::Get_LL(int len)
 	    if(FAmode==FA::QUOTE_CMD)
 	       FAmode=FA::RETRIEVE;
 
-	    xstring& loc=xstring::get_tmp(loc_c);
+	    xstring loc(loc_c);
 	    session->Close(); // loc_c is no longer valid.
 	    loc_c=0;
 
@@ -1159,8 +1228,7 @@ int FileCopyPeerFA::Get_LL(int len)
 	    size=NO_SIZE_YET;
 	    date=NO_DATE_YET;
 
-	    try_time=NO_DATE;
-	    retries=-1;
+	    upload_state.Clear();
 	    current->Timeout(0); // retry with new location.
 	    return 0;
 	 }
@@ -1168,13 +1236,23 @@ int FileCopyPeerFA::Get_LL(int len)
       SetError(session->StrError(res));
       return -1;
    }
-   if(res==0)
+   else if(res==0)
    {
+      debug((10,"copy-peer: EOF on %s\n",session->GetFileURL(session->GetFile())));
       eof=true;
       FileAccess::cache->Add(session,file,FAmode,FA::OK,this);
       SetSuggestedFileName(session->GetSuggestedFileName());
       session->Close();
    }
+   else if(res<=MAX_READ_TO_DELAY)
+   {
+      if(get_delay<=MAX_DELAY-DELAY_STEP)
+	 get_delay+=DELAY_STEP;
+      get_ll_timer.SetMicroSeconds(get_delay);
+      session->SuspendSlave();
+   }
+   else if(res>MAX_READ_TO_DELAY && get_delay>=DELAY_STEP)
+      get_delay-=DELAY_STEP;
    return res;
 }
 
@@ -1197,14 +1275,7 @@ int FileCopyPeerFA::Put_LL(const char *buf,int len)
 	 return 0;
       if(res==FA::STORE_FAILED)
       {
-	 try_time=session->GetTryTime();
-	 retries=session->GetRetries();
-	 off_t pos=session->GetRealPos();
-	 if(upload_watermark<pos) {
-	    upload_watermark=pos;
-	    retries=-1;
-	 }
-	 Log::global->Format(10,"try_time=%ld, retries=%d\n",(long)try_time,retries);
+	 upload_state.Save(session);
 	 session->Close();
 	 if(can_seek && seek_pos>0)
 	    Seek(FILE_END);
@@ -1256,15 +1327,15 @@ off_t FileCopyPeerFA::GetRealPos()
 
 void FileCopyPeerFA::Init()
 {
+   get_delay=0;
    fxp=false;
-   try_time=NO_DATE;
-   retries=-1;
-   upload_watermark=0;
    redirections=0;
    can_seek=true;
    can_seek0=true;
    if(FAmode==FA::LIST || FAmode==FA::LONG_LIST)
       Save(FileAccess::cache->SizeLimit());
+   if(mode==PUT)
+      file.set(UseTempFile(file));
 }
 
 FileCopyPeerFA::FileCopyPeerFA(FileAccess *s,const char *f,int m)
@@ -1324,12 +1395,12 @@ FileCopyPeer *FileCopyPeerFA::Clone()
 #define super FileCopyPeer
 
 FileCopyPeerFDStream::FileCopyPeerFDStream(FDStream *o,dir_t m)
-   : FileCopyPeer(m), my_stream(o?o:new FDStream(1,"<stdout>")), stream(my_stream)
+   : FileCopyPeer(m), my_stream(o?o:new FDStream(1,"<stdout>")), stream(my_stream), close_when_done(o!=0)
 {
    Init();
 }
 FileCopyPeerFDStream::FileCopyPeerFDStream(const Ref<FDStream>& o,dir_t m)
-   : FileCopyPeer(m), stream(o)
+   : FileCopyPeer(m), stream(o), close_when_done(false)
 {
    Init();
 }
@@ -1354,6 +1425,8 @@ void FileCopyPeerFDStream::Init()
       write_allowed=false;
    if(mode==PUT)
       put_ll_timer=new Timer(0,200);
+   if(mode==PUT && stream->fd==-1 && stream->can_setmtime())
+      stream->full_name.set(UseTempFile(stream->full_name));
 }
 
 void FileCopyPeerFDStream::Seek_LL()
@@ -1443,10 +1516,15 @@ int FileCopyPeerFDStream::Do()
 	 {
 	    const char *new_name=dir_file(dirname(stream->full_name),suggested_filename);
 	    struct stat st;
-	    if(lstat(new_name,&st)==-1 && errno==ENOENT) {
+	    if((lstat(new_name,&st)==-1 && errno==ENOENT) || ResMgr::QueryBool("xfer:clobber",0)) {
 	       debug((5,"copy: renaming `%s' to `%s'\n",stream->full_name.get(),suggested_filename.get()));
-	       if(rename(stream->full_name,new_name)==-1)
-		  debug((3,"rename(%s, %s): %s\n",stream->full_name.get(),new_name,strerror(errno)));
+	       if(rename(stream->full_name,new_name)==-1) {
+		  const char *err=xstring::format("rename(%s, %s): %s\n",stream->full_name.get(),new_name,strerror(errno));
+		  if(temp_file)
+		     SetError(err);
+		  else
+		     debug((3,"%s\n",err));
+	       }
 	    }
 	 }
 	 done=true;
@@ -1460,6 +1538,7 @@ int FileCopyPeerFDStream::Do()
    if(ascii)
       check_min_size=false;
 #endif
+   int res;
    switch(mode)
    {
    case PUT:
@@ -1467,7 +1546,9 @@ int FileCopyPeerFDStream::Do()
       {
 	 if(eof)
 	 {
-	    getfd(); // give it a chance to create empty file. (FIXME - handle tmp errors)
+	    // make sure the stream is open - it may create an empty file.
+	    if(stream && !stream->is_closed() && getfd()==-1)
+	       return m;
 	    if(!date_set && date!=NO_DATE && do_set_date)
 	    {
 	       if(date==NO_DATE_YET)
@@ -1476,7 +1557,7 @@ int FileCopyPeerFDStream::Do()
 	       date_set=true;
 	       m=MOVED;
 	    }
-	    if(stream && my_stream && !stream->Done())
+	    if(stream && close_when_done && !stream->Done())
 	       return m;
 	    if(!verify && do_verify)
 	       verify=new FileVerificator(stream);
@@ -1491,43 +1572,29 @@ int FileCopyPeerFDStream::Do()
 	 return m;
       if(getfd()==-1)
 	 return m;
-      while(Size()>0)
-      {
-	 if(check_min_size && !eof && Size()<PUT_LL_MIN
-	 && put_ll_timer && !put_ll_timer->Stopped())
-	    break;
-	 int res=Put_LL(buffer+buffer_ptr,Size());
-	 if(res>0)
-	 {
-	    buffer_ptr+=res;
-	    m=MOVED;
-	 }
-	 if(res<0)
-	    return MOVED;
-	 if(res==0)
-	    break;
-      }
+
+      if(check_min_size && !eof && Size()<PUT_LL_MIN
+      && put_ll_timer && !put_ll_timer->Stopped())
+	 break;
+      res=Put_LL(buffer+buffer_ptr,Size());
+      if(res>0)
+	 buffer_ptr+=res;
+      if(res!=0)
+	 m=MOVED;
       break;
 
    case GET:
       if(eof)
 	 return m;
-      while(Size()<GET_BUFSIZE)
+
+      res=TuneGetSize(Get_LL(get_size));
+      if(res>0)
       {
-	 int res=Get_LL(GET_BUFSIZE);
-	 if(res>0)
-	 {
-	    EmbraceNewData(res);
-	    SaveMaxCheck(0);
-	    m=MOVED;
-	 }
-	 if(res<0)
-	    return MOVED;
-	 if(eof)
-	    return MOVED;
-	 if(res==0)
-	    break;
+	 EmbraceNewData(res);
+	 SaveMaxCheck(0);
       }
+      if(res!=0 || eof)
+	 m=MOVED;
       break;
    }
    return m;
@@ -1558,7 +1625,9 @@ void FileCopyPeerFDStream::Seek(off_t new_pos)
    {
       if(seek_pos!=FILE_END)
       {
-	 pos=seek_pos+(mode==PUT?Size():0);
+	 pos=seek_pos;
+	 if(mode==PUT)
+	    pos+=Size();
 	 return;
       }
       else
@@ -1567,7 +1636,7 @@ void FileCopyPeerFDStream::Seek(off_t new_pos)
 	 if(s!=-1)
 	 {
 	    SetSize(s);
-	    pos=seek_pos+(mode==PUT)?Size():0;
+	    pos=seek_pos+((mode==PUT)?Size():0);
 	    return;
 	 }
 	 else
@@ -1646,8 +1715,10 @@ int FileCopyPeerFDStream::Get_LL(int len)
    }
 #endif
 
-   if(res==0)
+   if(res==0) {
+      debug((10,"copy-peer: EOF on FD %d\n",fd));
       eof=true;
+   }
    return res;
 }
 
@@ -1878,7 +1949,8 @@ int FileCopyPeerMemory::Do()
 void FileVerificator::Init0()
 {
    done=false;
-   if(!ResMgr::QueryBool("xfer:verify",0))
+   if(!ResMgr::QueryBool("xfer:verify",0)
+   || ResMgr::Query("xfer:verify-command",0).is_empty())
       done=true;
 }
 void FileVerificator::InitVerify(const char *f)
