@@ -1,7 +1,7 @@
 /*
  * lftp - file transfer program
  *
- * Copyright (c) 1996-2011 by Alexander V. Lukyanov (lav@yars.free.net)
+ * Copyright (c) 1996-2012 by Alexander V. Lukyanov (lav@yars.free.net)
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -23,7 +23,7 @@
 
 class FinderJob_Du : public FinderJob
 {
-   Ref<IOBuffer> buf;
+   SMTaskRef<IOBuffer> buf;
 
    /* We keep traversing deeper than this, but we never print a total
     * past this. */
