@@ -74,15 +74,6 @@ class MirrorJob : public Job
 	 parent_mirror->AddBytesToTransfer(b);
    }
 
-   const xstring& GetCmdLine() const {
-      if(bytes_to_transfer>0) {
-	 return xstring::get_tmp(cmdline).appendf(" - %lld/%lld (%d%%)",
-	    bytes_transferred,bytes_to_transfer,
-	    percent(bytes_transferred,bytes_to_transfer));
-      }
-      return cmdline;
-   }
-
    void	 HandleFile(FileInfo *);
 
    bool create_target_dir;
@@ -111,6 +102,8 @@ class MirrorJob : public Job
    };
    Statistics stats;
 
+   double transfer_time_elapsed;
+   TimeDate transfer_start_ts;
    int	 root_transfer_count;
    int	 &transfer_count;
 
@@ -126,6 +119,7 @@ class MirrorJob : public Job
    void	 va_Report(const char *fmt,va_list v);
    int	 verbose_report;
    MirrorJob *parent_mirror;
+   MirrorJob *root_mirror;
 
    time_t newer_than;
    time_t older_than;
@@ -152,9 +146,25 @@ class MirrorJob : public Job
    int target_redirections;
 
    void HandleChdir(FileAccessRef& session, int &redirections);
-   void HandleListInfoCreation(const FileAccessRef& session,Ref<ListInfo>& list_info,
+   void HandleListInfoCreation(const FileAccessRef& session,SMTaskRef<ListInfo>& list_info,
 	    const char *relative_dir);
-   void HandleListInfo(Ref<ListInfo>& list_info,Ref<FileSet>& set);
+   void HandleListInfo(SMTaskRef<ListInfo>& list_info,Ref<FileSet>& set);
+
+   void TransferStarted(class CopyJob *cp);
+   void JobStarted(Job *j);
+   void TransferFinished(Job *j);
+   void JobFinished(Job *j);
+
+   off_t GetBytesCount() {
+      long long bytes_count=Job::GetBytesCount();
+      if(!parent_mirror) {
+	 // bytes_transferred is cumulative over the mirror tree,
+	 // add it on the top only
+	 bytes_count+=bytes_transferred;
+      }
+      return bytes_count;
+   }
+   double GetTimeSpent() { return transfer_time_elapsed+(transfer_count>0?now-root_mirror->transfer_start_ts:0.); }
 
 public:
    enum
@@ -199,6 +209,7 @@ public:
    int	 Done() { return state==DONE; }
    void	 ShowRunStatus(const SMTaskRef<StatusLine>&);
    xstring& FormatStatus(xstring&,int v,const char *);
+   xstring& FormatShortStatus(xstring&);
    void	 SayFinal() { PrintStatus(0,""); }
    int	 ExitCode() { return stats.error_count!=0; }
 
