@@ -1,7 +1,7 @@
 /*
  * lftp - file transfer program
  *
- * Copyright (c) 1996-2011 by Alexander V. Lukyanov (lav@yars.free.net)
+ * Copyright (c) 1996-2012 by Alexander V. Lukyanov (lav@yars.free.net)
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -17,13 +17,12 @@
  * along with this program.  If not, see <http://www.gnu.org/licenses/>.
  */
 
-/* $Id: ResMgr.cc,v 1.82 2011/05/10 05:53:20 lav Exp $ */
-
 #include <config.h>
 
 #include <fnmatch.h>
 #include <ctype.h>
 #include <unistd.h>
+#include <stdlib.h>
 #include <math.h>
 #include <sys/types.h>
 #include <sys/stat.h>
@@ -39,11 +38,12 @@ CDECL_END
 #include "SMTask.h"
 #include "misc.h"
 #include "StringSet.h"
+#include "log.h"
 
-ResMgr::Resource  *ResMgr::chain=0;
-ResType		  *ResMgr::type_chain=0;
+xlist_head<Resource> Resource::all_list;
+xmap<ResType*> *ResType::types_by_name;
 
-int ResMgr::VarNameCmp(const char *good_name,const char *name)
+int ResType::VarNameCmp(const char *good_name,const char *name)
 {
    int res=EXACT_PREFIX+EXACT_NAME;
    const char *colon=strchr(good_name,':');
@@ -78,16 +78,18 @@ int ResMgr::VarNameCmp(const char *good_name,const char *name)
    return res;
 }
 
-const char *ResMgr::FindVar(const char *name,const ResType **type)
+const char *ResType::FindVar(const char *name,const ResType **type)
 {
    const ResType *exact_proto=0;
    const ResType *exact_name=0;
 
-   *type=0;
+   *type=types_by_name->lookup(name);
+   if(*type)
+      return 0;	// exact match
 
    int sub=0;
    const ResType *type_scan;
-   for(type_scan=type_chain; type_scan; type_scan=type_scan->next)
+   for(type_scan=types_by_name->each_begin(); type_scan; type_scan=types_by_name->each_next())
    {
       switch(VarNameCmp(type_scan->name,name))
       {
@@ -124,7 +126,7 @@ const char *ResMgr::FindVar(const char *name,const ResType **type)
    return _("ambiguous variable name");
 }
 
-const ResType *ResMgr::FindRes(const char *name)
+const ResType *ResType::FindRes(const char *name)
 {
    const ResType *type;
    const char *msg=FindVar(name,&type);
@@ -133,54 +135,48 @@ const ResType *ResMgr::FindRes(const char *name)
    return type;
 }
 
-const char *ResMgr::Set(const char *name,const char *cclosure,const char *cvalue)
+const char *ResType::Set(const char *name,const char *cclosure,const char *cvalue)
 {
-   const char *msg;
-
-   const ResType *type;
+   ResType *type;
    // find type of given variable
-   msg=FindVar(name,&type);
+   const char *msg=FindVar(name,&type);
    if(msg)
       return msg;
 
+   return type->Set(cclosure,cvalue);
+}
+
+const char *ResType::Set(const char *cclosure,const char *cvalue)
+{
+   const char *msg=0;
+
    xstring_c value(cvalue);
-   if(value && type->val_valid && (msg=(*type->val_valid)(&value))!=0)
+   if(value && val_valid && (msg=val_valid(&value))!=0)
       return msg;
 
    xstring_c closure(cclosure);
-   if(closure && type->closure_valid && (msg=(*type->closure_valid)(&closure))!=0)
+   if(closure && closure_valid && (msg=closure_valid(&closure))!=0)
       return msg;
 
-   Resource **scan;
-   // find the old value
-   for(scan=&chain; *scan; scan=&(*scan)->next)
-      if((*scan)->type==type
-	 && ((closure==0 && (*scan)->closure==0)
-	     || (closure && (*scan)->closure
-	         && !strcmp((*scan)->closure,closure))))
-	 break;
+   bool need_reconfig=false;
 
-   // if found
-   if(*scan)
+   xlist_for_each(Resource,*(type_value_list),node,scan)
    {
-      if(value)
-	 (*scan)->value.set(value);
-      else
+      // find the old value
+      if(closure==scan->closure || !xstrcmp(scan->closure,closure))
       {
-	 Resource *to_free=*scan;
-	 *scan=(*scan)->next;
-	 delete to_free;
-      }
-      ResClient::ReconfigAll(type->name);
-   }
-   else
-   {
-      if(value)
-      {
-	 chain=new Resource(chain,type,closure,value);
-	 ResClient::ReconfigAll(type->name);
+	 need_reconfig=true;
+	 delete scan;
+	 break;
       }
    }
+   if(value)
+   {
+      (void)new Resource(this,closure,value);
+      need_reconfig=true;
+   }
+   if(need_reconfig)
+      ResClient::ReconfigAll(name);
    return 0;
 }
 
@@ -198,163 +194,96 @@ int ResMgr::ResourceCompare(const Resource *ar,const Resource *br)
    return strcmp(ar->closure,br->closure);
 }
 
-int ResMgr::VResourceCompare(const void *a,const void *b)
+void Resource::Format(xstring& buf) const
 {
-   const ResMgr::Resource *ar=*(const ResMgr::Resource*const*)a;
-   const ResMgr::Resource *br=*(const ResMgr::Resource*const*)b;
-   return ResMgr::ResourceCompare(ar,br);
-}
-
-char *ResMgr::Format(bool with_defaults,bool only_defaults)
-{
-   Resource *scan;
-   ResType  *dscan;
-
-   int n=0;
-   int dn=0;
-   int size=0;
-   if(!only_defaults)
+   buf.appendf("set %s",type->name);
+   const char *s=closure;
+   if(s)
    {
-      for(scan=chain; scan; scan=scan->next)
-      {
-	 size+=4+strlen(scan->type->name);
-	 if(scan->closure)
-	    size+=1+1+2*strlen(scan->closure)+1;
-	 size+=1+1+2*strlen(scan->value)+1+1;
-	 n++;
-      }
-   }
-   if(with_defaults || only_defaults)
-   {
-      for(dscan=type_chain; dscan; dscan=dscan->next)
-      {
-	 size+=4+strlen(dscan->name);
-	 size+=1+1+2*xstrlen(dscan->defvalue)+1+1;
-	 dn++;
-      }
-   }
-
-   xstring res("");
-
-   Resource **created=(Resource**)alloca((dn+1)*sizeof(Resource*));
-   Resource **c_store=created;
-   dn=0;
-   if(with_defaults || only_defaults)
-   {
-      for(dscan=type_chain; dscan; dscan=dscan->next)
-      {
-	 if(only_defaults || SimpleQuery(dscan->name,0)==0)
-	 {
-	    dn++;
-	    *c_store++=new Resource(0,dscan,
-	       0,xstrdup(dscan->defvalue?dscan->defvalue:"(nil)"));
-	 }
-      }
-   }
-
-   Resource **arr=(Resource**)alloca((n+dn)*sizeof(Resource*));
-   n=0;
-   if(!only_defaults)
-   {
-      for(scan=chain; scan; scan=scan->next)
-	 arr[n++]=scan;
-   }
-   int i;
-   if(with_defaults || only_defaults)
-   {
-      for(i=0; i<dn; i++)
-	 arr[n++]=created[i];
-   }
-
-   qsort(arr,n,sizeof(*arr),&ResMgr::VResourceCompare);
-
-   for(i=0; i<n; i++)
-   {
-      res.appendf("set %s",arr[i]->type->name);
-      const char *s=arr[i]->closure;
-      if(s)
-      {
-	 res.append('/');
-	 bool par=false;
-	 if(strcspn(s," \t>|;&")!=strlen(s))
-	    par=true;
-	 if(par)
-	    res.append('"');
-	 while(*s)
-	 {
-	    if(strchr("\"\\",*s))
-	       res.append('\\');
-	    res.append(*s++);
-	 }
-	 if(par)
-	    res.append('"');
-      }
-      res.append(' ');
-      s=arr[i]->value;
-
+      buf.append('/');
       bool par=false;
-      if(*s==0 || strcspn(s," \t>|;&")!=strlen(s))
+      if(strcspn(s," \t>|;&")!=strlen(s))
 	 par=true;
       if(par)
-	 res.append('"');
+	 buf.append('"');
       while(*s)
       {
 	 if(strchr("\"\\",*s))
-	    res.append('\\');
-	 res.append(*s++);
+	    buf.append('\\');
+	 buf.append(*s++);
       }
       if(par)
-	 res.append('"');
-      res.append('\n');
+	 buf.append('"');
    }
-   for(i=0; i<dn; i++)
-      delete created[i];
-   return res.borrow();
+   buf.append(' ');
+   s=value;
+
+   bool par=false;
+   if(*s==0 || strcspn(s," \t>|;&")!=strlen(s))
+      par=true;
+   if(par)
+      buf.append('"');
+   while(*s)
+   {
+      if(strchr("\"\\",*s))
+	 buf.append('\\');
+      buf.append(*s++);
+   }
+   if(par)
+      buf.append('"');
+   buf.append('\n');
 }
 
-char **ResMgr::Generator(void)
+static int PResourceCompare(const Resource *const*a,const Resource *const*b)
 {
-   Resource *scan;
-   ResType  *dscan;
+   return ResMgr::ResourceCompare(*a,*b);
+}
+static int RefResourceCompare(const Ref<Resource> *a,const Ref<Resource> *b)
+{
+   return ResMgr::ResourceCompare(*a,*b);
+}
 
-   int n=0;
-   int dn=0;
-   for(scan=chain; scan; scan=scan->next)
-      n++;
-   for(dscan=type_chain; dscan; dscan=dscan->next)
-      dn++;
-
-   StringSet res;
-
-   Resource **created=(Resource**)alloca((dn+1)*sizeof(Resource*));
-   Resource **c_store=created;
-   dn=0;
-   for(dscan=type_chain; dscan; dscan=dscan->next)
+char *ResType::Format(bool with_defaults,bool only_defaults)
+{
+   RefArray<Resource> created;
+   if(with_defaults || only_defaults)
    {
-      if(SimpleQuery(dscan->name,0)==0)
-      {
-         dn++;
-	 *c_store++=new Resource(0,dscan,
-	     0,xstrdup(dscan->defvalue?dscan->defvalue:"(nil)"));
-      }
+      for(ResType *dscan=types_by_name->each_begin(); dscan; dscan=types_by_name->each_next())
+	 if(only_defaults || dscan->SimpleQuery(0)==0)
+	    created.append(new Resource(dscan,
+	       0,xstrdup(dscan->defvalue?dscan->defvalue:"(nil)")));
    }
 
-   Resource **arr=(Resource**)alloca((n+dn)*sizeof(Resource*));
-   n=0;
-   for(scan=chain; scan; scan=scan->next)
-      arr[n++]=scan;
+   xstring buf("");
 
-   int i;
-   for(i=0; i<dn; i++)
-      arr[n++]=created[i];
+   if(!only_defaults)
+   {
+      // just created Resources are also in all_list.
+      xarray<const Resource*> arr;
+      xlist_for_each(Resource,Resource::all_list,node,scan)
+	 arr.append(scan);
+      arr.qsort(PResourceCompare);
+      for(int i=0; i<arr.count(); i++)
+	 arr[i]->Format(buf);
+   }
+   else
+   {
+      created.qsort(RefResourceCompare);
+      for(int i=0; i<created.count(); i++)
+	 created[i]->Format(buf);
+   }
 
-   qsort(arr,n,sizeof(*arr),&ResMgr::VResourceCompare);
+   return buf.borrow();
+}
 
-   for(i=0; i<n; i++)
-      res.Append(arr[i]->type->name);
+char **ResType::Generator(void)
+{
+   StringSet res;
 
-   for(i=0; i<dn; i++)
-      delete created[i];
+   for(ResType *dscan=types_by_name->each_begin(); dscan; dscan=types_by_name->each_next())
+      res.Append(dscan->name);
+
+   res.qsort();
    return res.borrow();
 }
 
@@ -426,7 +355,7 @@ static const char power_letter[] =
 static unsigned long long get_power_multiplier(char p)
 {
    const char *scan=power_letter;
-   int scale=1024;
+   const int scale=1024;
    unsigned long long mul=1;
    p=toupper(p);
    while(scan<power_letter+sizeof(power_letter)) {
@@ -529,15 +458,19 @@ bool ResValue::to_tri_bool(bool a) const
    return to_bool();
 }
 
-ResMgr::Resource::Resource(Resource *next,const ResType *type,const char *closure,const char *value)
-   : type(type), value(value), closure(closure), next(next)
+Resource::Resource(ResType *type,const char *closure,const char *value)
+   : type(type), value(value), closure(closure), all_node(this), type_value_node(this)
 {
+   all_list.add_tail(all_node);
+   type->type_value_list->add_tail(type_value_node);
 }
-ResMgr::Resource::~Resource()
+Resource::~Resource()
 {
+   all_node.remove();
+   type_value_node.remove();
 }
 
-bool ResMgr::Resource::ClosureMatch(const char *cl_data)
+bool Resource::ClosureMatch(const char *cl_data)
 {
    if(!closure && !cl_data)
       return true;
@@ -557,43 +490,38 @@ bool ResMgr::Resource::ClosureMatch(const char *cl_data)
 
 const char *ResMgr::QueryNext(const char *name,const char **closure,Resource **ptr)
 {
-   const ResType *type=FindRes(name);
-   if(!type)
-      return 0;
-
+   xlist<Resource> *node=0;
    if(*ptr==0)
-      *ptr=chain;
-   else
-      *ptr=(*ptr)->next;
-
-   while(*ptr)
    {
-      if((*ptr)->type==type)
-      {
-	 *closure=(*ptr)->closure;
-	 return (*ptr)->value;
+      const ResType *type=FindRes(name);
+      if(!type) {
+	 *ptr=0;
+	 *closure=0;
+	 return 0;
       }
-      *ptr=(*ptr)->next;
+      node=type->type_value_list->get_next();
    }
+   else
+   {
+      node=(*ptr)->type_value_node.get_next();
+   }
+   *ptr=node->get_obj();
+   if(*ptr) {
+      *closure=(*ptr)->closure;
+      return (*ptr)->value;
+   }
+   *closure=0;
    return 0;
 }
 
-const char *ResMgr::SimpleQuery(const ResType *type,const char *closure)
+const char *ResType::SimpleQuery(const char *closure) const
 {
    // find the value
-   for(Resource *scan=chain; scan; scan=scan->next)
-      if(scan->type==type && scan->ClosureMatch(closure))
+   xlist_for_each(Resource,*(type_value_list),node,scan)
+      if(scan->ClosureMatch(closure))
 	 return scan->value;
    return 0;
 }
-const char *ResMgr::SimpleQuery(const char *name,const char *closure)
-{
-   const ResType *type=FindRes(name);
-   if(!type)
-      return 0;
-
-   return SimpleQuery(type,closure);
-}
 
 ResValue ResMgr::Query(const char *name,const char *closure)
 {
@@ -617,9 +545,9 @@ ResValue ResType::Query(const char *closure) const
    const char *v=0;
 
    if(closure)
-      v=ResMgr::SimpleQuery(this,closure);
+      v=SimpleQuery(closure);
    if(!v)
-      v=ResMgr::SimpleQuery(this,0);
+      v=SimpleQuery(0);
    if(!v)
       v=defvalue;
 
@@ -637,47 +565,53 @@ ResDecl::ResDecl(const char *a_name,const char *a_defvalue,ResValValid *a_val_va
    defvalue=a_defvalue;
    val_valid=a_val_valid;
    closure_valid=a_closure_valid;
-   ResMgr::AddType(this);
+   Register();
 }
 ResDecls::ResDecls(ResType *array)
 {
    while(array->name)
-      ResMgr::AddType(array++);
+      array++->Register();
 }
 ResDecls::ResDecls(ResType *r1,ResType *r2,...)
 {
-   ResMgr::AddType(r1);
+   r.append(r1);
+   r1->Register();
    if(!r2)
       return;
-   ResMgr::AddType(r2);
+   r.append(r2);
+   r2->Register();
    va_list v;
    va_start(v,r2);
    while((r1=va_arg(v,ResType *))!=0)
-      ResMgr::AddType(r1);
+   {
+      r1->Register();
+      r.append(r1);
+   }
    va_end(v);
 }
-
-ResType::~ResType()
+ResDecls::~ResDecls()
 {
-   for(ResType **scan=&ResMgr::type_chain; *scan; scan=&(*scan)->next)
-   {
-      if(*scan==this)
-      {
-	 *scan=this->next;
-	 break;
-      }
-   }
+   for(int i=0; i<r.count(); i++)
+      r[i]->Unregister();
+}
 
-   {
+void ResType::Register()
+{
+   if(!types_by_name)
+      types_by_name=new xmap<ResType*>;
+   types_by_name->add(name,this);
+   if(!type_value_list)
+      type_value_list=new xlist_head<Resource>();
+}
+void ResType::Unregister()
+{
+   types_by_name->remove(name);
+   if(type_value_list) {
       // remove all resources of this type
-      ResMgr::Resource **scan=&ResMgr::chain;
-      while(*scan)
-      {
-	 if((*scan)->type==this)
-	    delete replace_value(*scan,(*scan)->next);
-	 else
-	    scan=&(*scan)->next;
-      }
+      xlist_for_each_safe(Resource,*type_value_list,node,scan,next)
+	 delete scan;
+      delete type_value_list;
+      type_value_list=0;
    }
 }
 
@@ -734,71 +668,75 @@ const char *ResMgr::TimeIntervalValidate(xstring_c *s)
    return 0;
 }
 
-Range::Range(const char *s)
+void NumberPair::init(char sep1,const char *s)
 {
-   start=end=0;
-   no_start=no_end=true;
+   sep=sep1;
+   Set(s);
+}
+long long NumberPair::parse1(const char *s)
+{
+   if(!s || !*s)
+      return 0;
+   const char *end=s;
+   long long v=strtoll(s,const_cast<char**>(&end),0);
+   long long m=get_power_multiplier(*end);
+   if(s==end || m==0 || end[m>1]) {
+      error_text=_("invalid number");
+      return 0;
+   }
+   long long vm=v*m;
+   if(vm/m!=v) {
+      error_text=_("integer overflow");
+      return 0;
+   }
+   return vm;
+}
+void NumberPair::Set(const char *s0)
+{
+   n1=n2=0;
+   no_n1=no_n2=true;
    error_text=0;
 
-   if(!strcasecmp(s,"full") || !strcasecmp(s,"any"))
+   if(!s0)
       return;
 
-   int n;
-   int len=strlen(s);
-   char start_k=0,end_k=0;
-   if(sscanf(s,"%lld-%n",&start,&n)==1 && n==len)
-      no_start=false;
-   else if(sscanf(s,"-%lld%n",&end,&n)==1 && n==len)
-      no_end=false;
-   else if(sscanf(s,"%lld-%lld%n",&start,&end,&n)==2 && n==len)
-      no_start=no_end=false;
-   else if(sscanf(s,"%lld%c-%n",&start,&start_k,&n)==2 && n==len)
-      no_start=false;
-   else if(sscanf(s,"-%lld%c%n",&end,&end_k,&n)==2 && n==len)
-      no_end=false;
-   else if(sscanf(s,"%lld%c-%lld%n",&start,&start_k,&end,&n)==3 && n==len)
-      no_start=no_end=false;
-   else if(sscanf(s,"%lld-%lld%c%n",&start,&end,&end_k,&n)==3 && n==len)
-      no_start=no_end=false;
-   else if(sscanf(s,"%lld%c-%lld%c%n",&start,&start_k,&end,&end_k,&n)==4 && n==len)
-      no_start=no_end=false;
+   char *s1=alloca_strdup(s0);
+   char *s2=s1;
+   while(*s2 && *s2!=sep && *s2!=':')
+      s2++;
+   if(*s2)
+      *s2++=0;
    else
-   {
-      error_text=_("Invalid range format. Format is min-max, e.g. 10-20.");
-      return;
+      s2=0;
+
+   n1=parse1(s1);
+   no_n1=!*s1;
+   n2=(s2?parse1(s2):n1);
+   no_n2=(s2 && !*s2);
+
+   if(!error_text) {
+      Log::global->Format(10,"%s translated to pair %lld%c%lld (%d,%d)\n",
+	 s0,n1,sep,n2,no_n1,no_n2);
    }
-   if(start_k)
-      error_text=scale(&start,start_k);
-   if(!error_text && end_k)
-      error_text=scale(&end,end_k);
-   if(!no_start && !no_end && start>end)
-      start=replace_value(end,start);
 }
 
-const char *Range::scale(long long *value,char suf)
+Range::Range(const char *s) : NumberPair('-')
 {
-   // kilo, Mega, Giga, Tera, Peta, Exa, Zetta, Yotta
-   static const char s[]="kMGTPEZY";
-   const char *match=strchr(s,suf);
-   if(!match)
-      return _("Invalid suffix. Valid suffixes are: k, M, G, T, P, E, Z, Y");
-   unsigned shift=10*(match-s+1);
-   if(((*value<<shift)>>shift)!=*value)
-      return _("Integer overflow");
-   *value<<=shift;
-   return 0;
+   if(!strcasecmp(s,"full") || !strcasecmp(s,"any"))
+      return;
+   Set(s);
 }
 
 long long Range::Random()
 {
    random_init();
 
-   if(no_start && no_end)
+   if(no_n1 && no_n2)
       return random();
-   if(no_end)
-      return start+random();
+   if(no_n2)
+      return n1+random();
 
-   return start + (long long)((end-start+1)*random01());
+   return n1 + (long long)((n2-n1+1)*random01());
 }
 
 const char *ResMgr::RangeValidate(xstring_c *s)
@@ -806,6 +744,9 @@ const char *ResMgr::RangeValidate(xstring_c *s)
    Range r(*s);
    if(r.Error())
       return r.ErrorText();
+   char *colon=strchr(s->get_non_const(),':');
+   if(colon)
+      *colon='-';
    return 0;
 }
 
@@ -928,36 +869,29 @@ const char *ResMgr::NoClosure(xstring_c *)
 
 const char *ResMgr::UNumberPairValidate(xstring_c *value)
 {
-   int n=0;
-
-   unsigned a,b;
-   if(2>sscanf(value->get(),"%u%*c%u%n",&a,&b,&n))
-   {
-      if(UNumberValidate(value))
-	 return _("invalid pair of numbers");
-      return 0;
-   }
-
-   value->truncate(n);
-
+   NumberPair pair(':',*value);
+   if(pair.Error())
+      return pair.ErrorText();
    return 0;
 }
 void ResValue::ToNumberPair(int &a,int &b) const
 {
-   switch(sscanf(s,"%d%*c%d",&a,&b))
-   {
-   case 0: a=0;
-   case 1: b=a;
+   NumberPair pair(':',s);
+   if(pair.Error()) {
+      a=b=0;
+   } else {
+      a=pair.N1();
+      b=pair.HasN2()?pair.N2():a;
    }
 }
 
-ResClient *ResClient::chain;
+xlist_head<ResClient> ResClient::list;
 ResValue ResClient::Query(const char *name,const char *closure) const
 {
    if(!strchr(name,':'))
    {
       const char *prefix=ResPrefix();
-      name=xstring::format("%s:%s",prefix,name);
+      name=xstring::cat(prefix,":",name,NULL);
       name=alloca_strdup(name);
    }
    if(!closure)
@@ -965,16 +899,17 @@ ResValue ResClient::Query(const char *name,const char *closure) const
    return ResMgr::Query(name,closure);
 }
 ResClient::ResClient()
+   : node(this)
 {
-   ListAdd(ResClient,chain,this,next);
+   list.add(node);
 }
 ResClient::~ResClient()
 {
-   ListDel(ResClient,chain,this,next);
+   node.remove();
 }
 void ResClient::ReconfigAll(const char *r)
 {
-   ListScan(ResClient,chain,next)
+   xlist_for_each(ResClient,list,node,scan)
       scan->Reconfig(r);
 }
 
