@@ -25,74 +25,83 @@
 #include <time.h>
 #include "TimeDate.h"
 #include "xstring.h"
+#include "xlist.h"
+#include "xmap.h"
 
 typedef const char *ResValValid(xstring_c *value);
 typedef const char *ResClValid(xstring_c *closure);
 
 class ResValue;
+class ResMgr;
+class Resource;
 
 struct ResType
 {
+   static bool class_inited;
+   static xmap<ResType*> *types_by_name;
+
    const char *name;
    const char *defvalue;
    ResValValid *val_valid;
    ResClValid *closure_valid;
-   ResType *next;
-   ~ResType();
+   xlist_head<Resource> *type_value_list;
 
+   const char *SimpleQuery(const char *closure) const;
    ResValue Query(const char *closure) const;
    bool QueryBool(const char *closure) const;
    bool QueryTriBool(const char *closure,bool a) const;
-};
 
-class ResMgr
-{
-   static bool class_inited;
-   friend class ResType;
-public:
-   class Resource
-   {
-      friend class ResMgr;
-      friend class ResType;
+   const char *Set(const char *cclosure,const char *cvalue);
+   static const char *Set(const char *name,const char *closure,const char *value);
 
-      const ResType *type;
-      xstring_c value;
-      xstring_c closure;
+   void Register();
+   void Unregister();
 
-      Resource *next;
+   static char *Format(bool with_defaults,bool only_defaults);
+   static char **Generator(void);
 
-      bool ClosureMatch(const char *cl_data);
-
-      Resource(Resource *next,const ResType *type,const char *closure,const char *value);
-      ~Resource();
-   };
-
-private:
-   static Resource *chain;
-   static ResType *type_chain;
-
-public:
-   static void AddType(ResType *t) { t->next=type_chain; type_chain=t; }
-   static const char *QueryNext(const char *name,const char **closure,Resource **ptr);
-   static const char *SimpleQuery(const ResType *type,const char *closure);
-   static const char *SimpleQuery(const char *name,const char *closure);
-   static ResValue Query(const char *name,const char *closure);
-   static bool QueryBool(const char *name,const char *closure);
-   static bool QueryTriBool(const char *name,const char *closure,bool a);
+   static void ClassInit();
 
    enum CmpRes {
       EXACT_PREFIX=0x00,SUBSTR_PREFIX=0x01,
       EXACT_NAME  =0x00,SUBSTR_NAME  =0x10,
       DIFFERENT=-1
    };
-
    static int VarNameCmp(const char *name1,const char *name2);
    static const char *FindVar(const char *name,const ResType **type);
+   static const char *FindVar(const char *name,ResType **type) { return FindVar(name,const_cast<const ResType **>(type)); }
    static const ResType *FindRes(const char *name);
-   static const char *Set(const char *name,const char *closure,const char *value);
+};
 
-   static char *Format(bool with_defaults,bool only_defaults);
-   static char **Generator(void);
+class Resource
+{
+   friend class ResMgr;
+   friend struct ResType;
+   static xlist_head<Resource> all_list;
+
+   const ResType *type;
+   xstring_c value;
+   xstring_c closure;
+
+   xlist<Resource> all_node;
+   xlist<Resource> type_value_node;
+
+   bool ClosureMatch(const char *cl_data);
+   void Format(xstring& buf) const;
+
+   Resource(ResType *type,const char *closure,const char *value);
+public:
+   ~Resource();
+};
+
+class ResMgr : public ResType
+{
+   ResMgr();
+public:
+   static const char *QueryNext(const char *name,const char **closure,Resource **ptr);
+   static ResValue Query(const char *name,const char *closure);
+   static bool QueryBool(const char *name,const char *closure);
+   static bool QueryTriBool(const char *name,const char *closure,bool a);
 
    static const char *BoolValidate(xstring_c *value);
    static const char *TriBoolValidate(xstring_c *value);
@@ -116,10 +125,7 @@ public:
    static const char *NoClosure(xstring_c *);
    static bool str2bool(const char *value);
 
-   static void ClassInit();
-
    static int ResourceCompare(const Resource *a,const Resource *b);
-   static int VResourceCompare(const void *a,const void *b);
 };
 
 class ResDecl : public ResType
@@ -127,12 +133,15 @@ class ResDecl : public ResType
 public:
    ResDecl(const char *a_name,const char *a_defvalue,
 	   ResValValid *a_val_valid,ResClValid *a_closure_valid=0);
+   ~ResDecl() { Unregister(); }
 };
 class ResDecls
 {
+   xarray<ResType*> r;
 public:
    ResDecls(ResType *array);
    ResDecls(ResType *r1,ResType *r2,...);
+   ~ResDecls();
 };
 
 class ResValue
@@ -161,6 +170,7 @@ public:
 	 return s;
       }
    bool is_nil() const { return s==0; }
+   bool is_empty() const { return s==0 || *s==0; }
    void ToNumberPair(int &a,int &b) const;
 };
 
@@ -214,8 +224,8 @@ public:
 
 class ResClient
 {
-   static ResClient *chain;
-   ResClient *next;
+   static xlist_head<ResClient> list;
+   xlist<ResClient> node;
 protected:
    virtual const char *ResPrefix() const { return 0; }
    virtual const char *ResClosure() const { return 0; }
