@@ -230,7 +230,7 @@ private:
    Ref<FileSet> file_set;
    Timer flush_timer;
 
-   void Disconnect();
+   void DisconnectLL();
    int IsConnected() const
       {
 	 if(state==DISCONNECTED)
@@ -772,7 +772,7 @@ public:
 
    int Do();
    int Done();
-   int Read(void *,int);
+   int Read(Buffer *,int);
    int Write(const void *,int);
    int StoreStatus();
    int Buffered();
@@ -802,7 +802,7 @@ public:
 
 class SFtpDirList : public DirList
 {
-   Ref<IOBuffer> ubuf;
+   SMTaskRef<IOBuffer> ubuf;
    const char *dir;
    bool use_file_set;
    Ref<FileSet> fset;
