@@ -40,6 +40,11 @@
 # include <sys/time.h>
 #endif
 
+#if LIBIDN
+# include <idna.h>
+# include <tld.h>
+#endif
+
 CDECL_BEGIN
 #include "regex.h"
 #include "human.h"
@@ -998,3 +1003,26 @@ const char *xhuman(long long n)
    char *buf=xstring::tmp_buf(LONGEST_HUMAN_READABLE + 1);
    return human_readable(n, buf, human_autoscale|human_SI, 1, 1);
 }
+
+const char *xidna_to_ascii(const char *name)
+{
+#if LIBIDN
+   if(!name)
+      return 0;
+   char *name_ace_tmp=0;
+   if(idna_to_ascii_lz(name,&name_ace_tmp,0)==IDNA_SUCCESS)
+      return xstring::get_tmp().set_allocated(name_ace_tmp);
+#endif//LIBIDN
+   return name;
+}
+bool xtld_name_ok(const char *name)
+{
+#if LIBIDN
+   if(mbswidth(name,MBSW_REJECT_INVALID|MBSW_REJECT_UNPRINTABLE)<=0)
+      return false;
+   size_t err_pos;
+   if(tld_check_lz(name,&err_pos,NULL)==TLD_SUCCESS)
+      return true;
+#endif//LIBIDN
+   return false;
+}
